'''
Utilities for a feature extraction
'''

import pickle
import os
from os.path import join
from io import StringIO
import collections as clctn

from tqdm import tqdm

from . import color as c
from . import textual_feature

def parse_tess(file_name):
	'''Used to parse tess tags found at the beginning of lines of .tess files'''
	file_text = StringIO()
	with open(file_name, mode='r', encoding='utf-8') as file:
		for line in file:
			#Ignore lines without tess tags, or parse the tag out and strip whitespace
			if not line.startswith('<'):
				continue
			assert '>' in line, f'Malformed tess tag in {file_name}'
			file_text.write(line[line.index('>') + 1:].strip())
			file_text.write(' ')
	return file_text.getvalue()

FILE_PARSERS = {
	'tess': parse_tess,
}

def _get_filenames(corpus_dir, file_extensions, excluded_paths):
	#Obtain all the files to parse by traversing through the corpus directory
	file_names = []
	for current_path, current_dir_names, current_file_names in os.walk(corpus_dir, topdown=True):
		#Remove the excluded directories to prevent traversing into them
		del_indexes = []
		for i, cur_dir in enumerate(current_dir_names):
			if join(current_path, cur_dir) + os.sep in excluded_paths:
				del_indexes.append(i)
		#Iterate backwards to prevent index removal issues
		for i in range(len(del_indexes) - 1, -1, -1):
			del current_dir_names[del_indexes[i]]

		for current_file_name in current_file_names:
			if '.' in current_file_name and current_file_name[current_file_name.rindex('.') + 1:] in file_extensions \
			and join(current_path, current_file_name) not in excluded_paths:
				file_names.append(join(current_path, current_file_name))
	return sorted(file_names)

def _extract_features(corpus_dir, file_extension_to_parse_function, excluded_paths, features, output_file):
	file_names = _get_filenames(corpus_dir, file_extension_to_parse_function.keys(), excluded_paths)
	feature_tuples = [(name, textual_feature.decorated_features[name]) for name in features]
	text_to_features = {} #Associates file names to their respective features
	print(
		f'Extracting features from file with extensions '
		f'[{", ".join(file_extension_to_parse_function.keys())}] in directory {c.yellow(corpus_dir)}'
	)

	#Feature extraction
	for file_name in file_names if output_file is None else tqdm(file_names, dynamic_ncols=True):
		text_to_features[file_name] = {}
		file_extension = file_name[file_name.rindex('.') + 1:]
		file_text = file_extension_to_parse_function[file_extension](file_name)

		for feature_name, feature_func in feature_tuples:
			score = feature_func(text=file_text, filepath=file_name)
			text_to_features[file_name][feature_name] = score
			if output_file is None:
				print(f'{file_name}, {str(feature_name)}, {c.green(str(score))}')

	textual_feature.clear_cache()

	if output_file is not None:
		print(f'Feature mining complete. Attempting to write feature results to "{c.yellow(output_file)}"...')
		with open(output_file, 'wb') as pickle_file:
			pickle_file.write(pickle.dumps(text_to_features))
		print(c.green('Success!'))

# Keys of file_extension_to_parse_function must not include the dot e.g. use txt not .txt
# If excluded_paths is given, it must be a set and it can contain files or directories (the directories must
# end in a file separator e.g. slash on Mac or Linux)
#pylint: disable = too-many-branches
def main(corpus_dir, file_extension_to_parse_function, excluded_paths=None, features=None, output_file=None):
	'''Run feature extraction on all decorated features'''
	if excluded_paths is None: excluded_paths = set()
	if features is None: features = textual_feature.decorated_features.keys()

	if not corpus_dir: raise ValueError('Must provide a directory that contains the corpus')
	if not file_extension_to_parse_function or not isinstance(file_extension_to_parse_function, clctn.Mapping):
		raise ValueError('Must provide a mapping from file extensions to functions specifying how to parse them')
	if None in file_extension_to_parse_function:
		raise ValueError('The keys of file_extension_to_parse_function must not be None')
	if not all(callable(f) for f in file_extension_to_parse_function.values()):
		raise ValueError('The values of file_extension_to_parse_function must be callable')
	if any('.' in extension for extension in file_extension_to_parse_function.keys()):
		raise ValueError(
			f'The following extensions contain a period. '
			f'Please remove the periods: {list(filter(lambda x: "." in x, file_extension_to_parse_function.keys()))}'
		)
	if not features:
		raise ValueError(
			f'No features were provided. Ensure you have declared and annotated '
			f'them with the decorator @textual_feature before calling this function'
		)
	if not os.path.isdir(corpus_dir): raise ValueError(f'Path "{corpus_dir}" is not a valid directory')
	if not isinstance(excluded_paths, set): raise ValueError('Excluded paths must be in a set')
	if not all(os.path.isfile(path) or os.path.isdir(path) for path in excluded_paths):
		raise ValueError(f'Each path in {str(excluded_paths)} must be a valid path for a file or directory!')
	if not all(name in textual_feature.decorated_features.keys() for name in features):
		raise ValueError(
			f'The values in set {str(set(features) - textual_feature.decorated_features.keys())} '
			f'are not among the decorated features in {str(textual_feature.decorated_features.keys())}'
		)

	if output_file:
		if not isinstance(output_file, str): raise ValueError('Output file must be a string for a file path')
		if os.path.isfile(output_file): raise ValueError(f'Output file "{output_file}" already exists!')
		if os.path.isdir(output_file):
			raise ValueError(f'The end of the path "{output_file}" is a directory - please specify a filename')
		if os.sep in output_file and not os.path.isdir(os.path.dirname(output_file)):
			raise ValueError(f'"{os.path.dirname(output_file)}" is not a valid directory!')
	elif output_file is not None: raise ValueError('Output file must be truthy, or None')

	from timeit import timeit
	from functools import partial
	print(
		'\n\n' + c.green(
			'Feature mining elapsed time: ' + '%.4f' % timeit(
				partial(
					_extract_features, corpus_dir, file_extension_to_parse_function,
					excluded_paths, features, output_file
				),
				number=1
			) + ' seconds'
		)
	)



#pylint: disable = protected-access, invalid-name, unnecessary-lambda
'''Utilities for textual feature decorator'''
import re
from inspect import signature
from collections import OrderedDict
from io import StringIO

import nltk.tokenize.punkt as punkt

decorated_features = OrderedDict()
lang = None
word_tokenizer = None
sentence_tokenizers = None
debug_output = StringIO()
NON_WORD_CHARS = (
	r"\?¿؟\!¡！‽…⋯᠁ฯ,،，､、。°※··᛫~\:;;\\\/⧸⁄（）\(\)\[\]\{\}\<\>"
	r"\'\"‘’“”‹›«»《》\|‖\=\-\‐\‒\–\—\―_\+\*\^\$£€§%#@&†‡"
)

tokenize_types = {
	None: {
		'func': lambda text: text,
		'prev_filepath': None,
		'tokens': None,
	},
	'sentences': {
		'func': lambda text: sentence_tokenizers[lang].tokenize(text),
		'prev_filepath': None,
		'tokens': None,
	},
	'words': {
		'func': lambda text: word_tokenizer.word_tokenize(text),
		'prev_filepath': None,
		'tokens': None,
	},
	'sentence_words': {
		'func': lambda text: [word_tokenizer.word_tokenize(s) for s in sentence_tokenizers[lang].tokenize(text)],
		'prev_filepath': None,
		'tokens': None,
	},
}

def clear_cache():
	'''Clear tokens from previously parsed texts'''
	global tokenize_types
	global debug_output
	for _, val in tokenize_types.items():
		val['prev_filepath'] = None
		val['tokens'] = None
	debug_output.truncate(0)
	debug_output.seek(0)

def setup_tokenizers(*, language=None, terminal_punctuation):
	'''Initialize the word tokenizer and sentence tokenizer given the terminal punctuation'''
	global lang
	global word_tokenizer
	global sentence_tokenizers
	global tokenize_types
	lang = language #TODO validate in this function, not in textual_feature()
	punkt.PunktLanguageVars.sent_end_chars = terminal_punctuation
	punkt.PunktLanguageVars.re_boundary_realignment = re.compile(r'[›»》’”\'\"）\)\]\}\>]+?(?:\s+|(?=--)|$)', re.MULTILINE)
	clear_cache()

	'''
	Accessing private variables of punkt.PunktLanguageVars because
	nltk has a faulty design pattern that necessitates it.
	Issue reported here: https://github.com/nltk/nltk/issues/2068
	'''

	'''
	A word tokenizer should strip the non word chars from words,
	as well as periods and numbers
	'''
	word_tokenizer = punkt.PunktLanguageVars()
	word_tokenizer._re_word_tokenizer = re.compile(punkt.PunktLanguageVars._word_tokenize_fmt % {
		'NonWord': fr"(?:[\d\.{NON_WORD_CHARS}])",
		'MultiChar': punkt.PunktLanguageVars._re_multi_char_punct,
		'WordStart': fr"[^\d\.{NON_WORD_CHARS}]",
	}, re.UNICODE | re.VERBOSE)
	word_tokenizer._re_period_context = re.compile(punkt.PunktLanguageVars._period_context_fmt % {
		'NonWord': fr"(?:[\d\.{NON_WORD_CHARS}])",
		'SentEndChars': word_tokenizer._re_sent_end_chars,
	}, re.UNICODE | re.VERBOSE)

	'''
	A sentence tokenizer should strip the non word chars from words.
	This regex excludes periods because the original regex in the Punkt class
	excludes them, and it excludes numbers because we do not want to
	treat numbers with decimals as if they were sentences
	'''
	sent_tok_vars = punkt.PunktLanguageVars()
	sent_tok_vars._re_word_tokenizer = re.compile(punkt.PunktLanguageVars._word_tokenize_fmt % {
		'NonWord': fr"(?:[{NON_WORD_CHARS}])",
		'MultiChar': punkt.PunktLanguageVars._re_multi_char_punct,
		'WordStart': fr"[^{NON_WORD_CHARS}]",
	}, re.UNICODE | re.VERBOSE)
	sent_tok_vars._re_period_context = re.compile(punkt.PunktLanguageVars._period_context_fmt % {
		'NonWord': fr"(?:[{NON_WORD_CHARS}])",
		'SentEndChars': sent_tok_vars._re_sent_end_chars,
	}, re.UNICODE | re.VERBOSE)

	sentence_tokenizers = {None: punkt.PunktSentenceTokenizer(lang_vars=punkt.PunktLanguageVars())}
	for sen_tkzr in sentence_tokenizers.values():
		sen_tkzr._lang_vars._re_period_context = sent_tok_vars._re_period_context
		sen_tkzr._lang_vars._re_word_tokenizer = sent_tok_vars._re_word_tokenizer

def textual_feature(*, tokenize_type=None, debug=False):
	'''Decorator for textual features'''
	global lang
	if not (word_tokenizer and sentence_tokenizers):
		raise ValueError(
			f'Tokenizers not initialized: Use '
			f'"setup_tokenizers(terminal_punctuation=<collection of punctutation>)" before decorating a function'
		)
	if tokenize_type not in tokenize_types:
		raise ValueError(
			'"' + str(tokenize_type) + '" is not a valid tokenize type: Choose from among ' +
			str(list(tokenize_types.keys()))
		)
	if lang not in sentence_tokenizers:
		raise ValueError(
			f'"{str(lang)}" is not an available language. Choose from among '
			f'{list(sentence_tokenizers.keys())}'
		)
	def decor(f):
		#TODO make this more extensible. Use keyword args somehow instead of 'text' parameter?
		#TODO Ensure that features with duplicated names aren’t put in the ordered dict (this can happen if they come from different files)
		reqrd_params = {'text'}
		sig_params = signature(f).parameters
		if not all(tok in sig_params for tok in reqrd_params):
			raise ValueError(
				f'Error for feature "{f.__name__}"'
				f'\nMinimal required parameters: {str(reqrd_params)}'
				f'\nFound parameters: {set(sig_params) if sig_params else "{}"}'
			)
		def wrapper(*, text, filepath=None):
			if not filepath:
				return f(tokenize_types[tokenize_type]['func'](text))

			#Cache the tokenized version of this text if this filepath is new
			if tokenize_types[tokenize_type]['prev_filepath'] != filepath:
				tokenize_types[tokenize_type]['prev_filepath'] = filepath
				tokenize_types[tokenize_type]['tokens'] = tokenize_types[tokenize_type]['func'](text)
			elif debug:
				debug_output.write('Cache hit! ' + 'function: <' + f.__name__ + '>, filepath: ' + filepath + '\n')
			return f(tokenize_types[tokenize_type]['tokens'])
		decorated_features[f.__name__] = wrapper
		return wrapper
	return decor



'''
Setup
'''
import setuptools
import pipfile

setuptools.setup(
	name='qcrit',
	version='0.0.14',
	author='Tim Gianitsos',
	author_email='contact@qcrit.org',
	description='Quantitative Criticism Lab',
	long_description=open('README.md', 'r').read(),
	long_description_content_type='text/markdown',
	url='https://www.qcrit.org',
	project_urls={'Source Code': 'https://github.com/QuantitativeCriticismLab/qcrit'},
	packages=setuptools.find_packages(),
	classifiers=[
		'Programming Language :: Python :: 3.6',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
	],
	install_requires=[
		pkg + (version if version != '*' else '')
		for pkg, version in pipfile.load('Pipfile').data['default'].items()
	],
)



'''
Analyze models
'''
from functools import partial
from collections import OrderedDict
import os
import pickle
import csv

import numpy as np

from . import model_analyzer
from . import color as c

def _get_features(feature_data_file):
	#Obtain features that were previously mined and serialized into a file
	filename_to_features = None
	with open(feature_data_file, mode='rb') as pickle_file:
		filename_to_features = pickle.loads(pickle_file.read())
	return filename_to_features

def _get_file_classifications(classification_data_file):
	#Obtain classifications for each file
	with open(classification_data_file, mode='r') as classification_file:
		csv_reader = csv.reader(classification_file)
		filename_to_classification = {}
		label_val_to_label_name = OrderedDict(
			(tok.split(':')[1], tok.split(':')[0])
			for tok in next(csv_reader)
		)
		next(csv_reader)
		for line in csv_reader:
			filename_to_classification[line[0]] = line[1]
		assert all(v in label_val_to_label_name for v in filename_to_classification.values())
	return filename_to_classification, label_val_to_label_name

def _get_classifier_data(filename_to_features, filename_to_classification, file_names, feature_names):
	data_1d = [filename_to_features[file_name][feature] for file_name in file_names for feature in feature_names]
	data = []
	for i in range(len(file_names)):
		data.append([val for val in data_1d[i * len(feature_names): i * len(feature_names) + len(feature_names)]])
	target = [filename_to_classification[file_name] for file_name in file_names]

	assert data[-1][-1] == data_1d[-1]
	assert len(data) == len(target)
	assert len(data) == len(file_names)
	assert len(feature_names) == len(data[0])

	#Convert lists to numpy arrays so they can be used in the machine learning models
	data = np.asarray(data)
	target = np.asarray(target)
	return (data, target)

#TODO unit test this
def main(feature_data_file, classification_data_file, model_funcs=None):
	'''Runs all decorated model analyzers'''

	if model_funcs is None: model_funcs = model_analyzer.DECORATED_ANALYZERS.keys()
	if not os.path.isfile(feature_data_file): raise ValueError('File "' + feature_data_file + '" does not exist')
	if not os.path.isfile(classification_data_file):
		raise ValueError('File "' + classification_data_file + '" does not exist')
	if not model_funcs: raise ValueError('No model analyzers were provided')
	if not all(f in model_analyzer.DECORATED_ANALYZERS for f in model_funcs):
		raise ValueError(
			'The values in set ' + str(set(model_funcs) - model_analyzer.DECORATED_ANALYZERS.keys()) +
			' are not among the decorated model analyzers in ' + str(model_analyzer.DECORATED_ANALYZERS.keys())
		)

	filename_to_features = _get_features(feature_data_file)

	filename_to_classification, label_val_to_label_name = _get_file_classifications(classification_data_file)

	if filename_to_features.keys() - filename_to_classification.keys():
		raise ValueError(
			'There exist some files in the feature_data_file for which no label exists in ' +
			'the classification_data_file: {\n\t' +
			'\n\t'.join(filename_to_features.keys() - filename_to_classification.keys()) + '\n}'
		)

	#Filter out unused labels (i.e. a label exists for a file with that name but no features were extracted for it)
	used_label_numbers = {filename_to_classification[filename] for filename in filename_to_features.keys()}
	label_val_to_label_name = OrderedDict(
		(k, v) for k, v in label_val_to_label_name.items() if k in used_label_numbers
	)

	#Convert features and classifications into sorted lists
	file_names = sorted([elem for elem in filename_to_features.keys()])
	feature_names = sorted(
		feature_name for feature_name in next(iter(filename_to_features.values())).keys()
	)

	data, target = _get_classifier_data(filename_to_features, filename_to_classification, file_names, feature_names)

	from timeit import timeit
	for funcname in model_funcs:
		print(
			'\n\n' + c.green(
				'Elapsed time: ' + '%.4f' % timeit(
					partial(
						model_analyzer.DECORATED_ANALYZERS[funcname], data, target, file_names,
						feature_names, label_val_to_label_name
					),
					number=1
				) + ' seconds'
			) + '\n'
		)



'''Decorator for functions analyzing models'''

from collections import OrderedDict

DECORATED_ANALYZERS = OrderedDict()

def model_analyzer():
	'''Decorator for functions analyzing models'''
	def decor(dec_func):
		def wrapper(data, target, file_names, feature_names, labels_key):
			return dec_func(data, target, file_names, feature_names, labels_key)
		DECORATED_ANALYZERS[dec_func.__name__] = wrapper
		return wrapper
	return decor



'''
ANSI Colors
'''
def red(input_str):
	'''Output red'''
	return '\033[91m' + input_str + '\033[0m'
def green(input_str):
	'''Output green'''
	return '\033[92m' + input_str + '\033[0m'
def yellow(input_str):
	'''Output yellow'''
	return '\033[93m' + input_str + '\033[0m'
def purple(input_str):
	'''Output purple'''
	return '\033[95m' + input_str + '\033[0m'
